/*
This is a part of the IconDesk source code.

Copyright (C) 2003 Erik Christiansson, aka Sci
www.alphafish.com
erik@alphafish.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "Group.h"

Group::Group(const char* nameIn, HWND parentWnd)
{
	name = (char*)malloc(strlen(nameIn) + 1);
	memcpy(name, nameIn, strlen(nameIn) + 1);
	_parentWnd = parentWnd;

	_iconsLoaded = FALSE;

	// Icon settings
	char buffer[MAX_LINE_LENGTH];

	_visible = _getBool("Visible", TRUE);
	_iconX = _getInt("X", 0);
	_iconY = _getInt("Y", 0);
	_iconSize = _getInt("Size", 32);
	_iconSpacingX = _getInt("SpacingX", 0);
	_iconSpacingY = _getInt("SpacingY", 0);

	FILE* fpStep;
	fpStep = LCOpen(NULL);
	if(fpStep) {
		char szPrefix[MAX_LINE_LENGTH], szBuffer[MAX_LINE_LENGTH], szData[MAX_LINE_LENGTH], szDummy[MAX_LINE_LENGTH];
		char* rszTokens[1];

		rszTokens[0] = szDummy;
		sprintf(szPrefix, "*IconDesk%sFolder", name);

		szBuffer[0] = 0;
		while(LCReadNextConfig(fpStep, szPrefix, szBuffer, sizeof(szBuffer))) {
			szDummy[0] = szData[0] = 0;
			LCTokenize(szBuffer, rszTokens, 1, szData);
			if(szData[strlen(szData) - 1] != '\\') {
				strcat(szData, "\\");
			}
			_folders.push_back(szData);

			// Folder watcher
			HANDLE hChange = FindFirstChangeNotification(szData, FALSE, FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME | FILE_NOTIFY_CHANGE_ATTRIBUTES);
			if(hChange == INVALID_HANDLE_VALUE) {
				LSLogPrintf(LOG_ERROR, "IconDesk", "Could not create change notification handle for '%s'", szData);
			}
			changeHandles.push_back(hChange);
		}
		LCClose(fpStep);
	}

	if(_folders.size() == 0) {
		_folders.push_back("C:\\");
		HANDLE hChange = FindFirstChangeNotification("C:\\", FALSE, FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME | FILE_NOTIFY_CHANGE_ATTRIBUTES);
		if(hChange == INVALID_HANDLE_VALUE) {
			LSLog(LOG_ERROR, "IconDesk", "Could not create change notification handle for 'C:\\'");
		}
		changeHandles.push_back(hChange);
	}

	_iconHideExtension = _getBool("HideExtension", TRUE);
	_iconFoldersFirst = _getBool("FoldersFirst", TRUE);
	_iconShowHidden = _getBool("ShowHidden", FALSE);

	_hideText = _getBool("HideText", FALSE);
	if(!_hideText) {
		_getString("TextBackground", buffer, "", MAX_LINE_LENGTH);
		if(strlen(buffer)) {
			_iconTextBackground = CreateSolidBrush(_getColor("TextBackground", RGB(255,255,255)));
		}
		else {
			_iconTextBackground = NULL;
		}
		_iconTextSpacing = _getInt("TextSpacing", 5);
		_iconTextHeight = _getInt("TextHeight", 28);
		_iconTextWidth = _getInt("TextWidth", 64);

		_iconTextColor = _getColor("FontColor", RGB(255, 255, 255));

		_getString("Font", buffer, "Arial", MAX_LINE_LENGTH);
		int fontSize = _getInt("FontSize", 14);
		BOOL fontItalic = _getBool("FontItalic", FALSE);
		BOOL fontBold = _getBool("FontBold", FALSE);
		if(!(_iconTextFont = CreateFont(fontSize, 0, 0, 0, fontBold ? FW_BOLD : FW_NORMAL, fontItalic, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, buffer))) {
			LSLogPrintf(LOG_ERROR, "IconDesk", "Could not load font \"%s\"", buffer);
		}

		_textShadowColor = _getColor("TextShadowColor", RGB(80, 80, 80));
		_textShadowDepth = _getInt("TextShadowDepth", 0);
	}
	else {
		_iconTextWidth = 0;
		_iconTextHeight = 0;
		_iconTextSpacing = 0;
	}

	_getString("Direction", buffer, "", MAX_LINE_LENGTH);
	if(!stricmp(buffer, "down")) {
		_iconDirection = DD_DOWN;
	}
	else if(!stricmp(buffer, "left")) {
		_iconDirection = DD_LEFT;
	}
	else if(!stricmp(buffer, "up")) {
		_iconDirection = DD_UP;
	}
	else {
		_iconDirection = DD_RIGHT;
	}

	_getString("WrapDirection", buffer, "", MAX_LINE_LENGTH);
	if(_iconDirection == DD_DOWN || _iconDirection == DD_UP) {
		if(!stricmp(buffer, "left")) {
			_iconWrapDirection = DD_LEFT;
		}
		else {
			_iconWrapDirection = DD_RIGHT;
		}
	}
	else {
		if(!stricmp(buffer, "up")) {
			_iconWrapDirection = DD_UP;
		}
		else {
			_iconWrapDirection = DD_DOWN;
		}
	}
	_iconWrapCount = max(_getInt("WrapCount", 10), 1);

	_getString("MouseLeftUp", _mouseLUp, ".none", MAX_PATH);
	_getString("MouseLeftDown", _mouseLDown, ".none", MAX_PATH);
	_getString("MouseLeftDbl", _mouseLDbl, ".open", MAX_PATH);
	_getString("MouseMiddleUp", _mouseMUp, ".none", MAX_PATH);
	_getString("MouseMiddleDown", _mouseMDown, ".none", MAX_PATH);
	_getString("MouseMiddleDbl", _mouseMDbl, ".none", MAX_PATH);
	_getString("MouseRightUp", _mouseRUp, ".context", MAX_PATH);
	_getString("MouseRightDown", _mouseRDown, ".none", MAX_PATH);
	_getString("MouseRightDbl", _mouseRDbl, ".none", MAX_PATH);

	// Icon effects
	_effectFlags = 0;
	_hueIntensity = (UCHAR)min(max(_getInt("HueIntensity", 0), 0), 255);
	_hueColor = _getColor("HueColor", RGB(128, 128, 128));
	if(_hueIntensity) {
		_effectFlags |= FX_HUE;
	}
	_saturation = (UCHAR)min(max(_getInt("Saturation", 255), 0), 255);
	if(_saturation != 255) {
		_effectFlags |= FX_SATURATION;
	}
}

Group::~Group()
{
	_deleteIcons();

	if(_iconTextFont) {
		DeleteObject(_iconTextFont);
	}
	if(_iconTextBackground) {
		DeleteObject(_iconTextBackground);
	}

	for(HandleListIter it = changeHandles.begin(); it != changeHandles.end(); it++) {
		FindCloseChangeNotification(*it);
	}

	free(name);
}

void Group::show()
{
	_visible = TRUE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Group::hide()
{
	_visible = FALSE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Group::toggle()
{
	_visible = !_visible;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Group::update()
{
	_deleteIcons();
	_loadIcons();
}

void Group::_mouseCommand(DESKTOPICON* clicked, const char* command)
{
	if(!stricmp(command, ".none")) {
		return;
	}
	else if(!stricmp(command, ".context")) {
		PopupContextMenu(clicked->filename);
	}
	else if(!stricmp(command, ".open")) {
		LSExecuteEx(NULL, NULL, clicked->filename, NULL, NULL, SW_NORMAL);
	}
}

BOOL Group::mouseEvent(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	DESKTOPICON *clicked = _getIconAtPt(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));

	if(clicked) {
		if(uMsg == WM_RBUTTONDOWN || uMsg == WM_LBUTTONDOWN || WM_MBUTTONDOWN) {
			if(_selectedIcon) {
				InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
			}
			_selectedIcon = clicked;
			if(_selectedIcon) {
				InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
			}
		}

		switch(uMsg) {
			case WM_RBUTTONUP: 
				_mouseCommand(clicked, _mouseRUp);
				break;

			case WM_RBUTTONDOWN:
				_mouseCommand(clicked, _mouseRDown);
				break;

			case WM_RBUTTONDBLCLK:
				_mouseCommand(clicked, _mouseRDbl);
				break;

			case WM_LBUTTONUP: 
				_mouseCommand(clicked, _mouseLUp);
				break;

			case WM_LBUTTONDOWN:
				_mouseCommand(clicked, _mouseLDown);
				break;

			case WM_LBUTTONDBLCLK:
				_mouseCommand(clicked, _mouseLDbl);
				break;

			case WM_MBUTTONUP: 
				_mouseCommand(clicked, _mouseMUp);
				break;

			case WM_MBUTTONDOWN:
				_mouseCommand(clicked, _mouseMDown);
				break;

			case WM_MBUTTONDBLCLK:
				_mouseCommand(clicked, _mouseMDbl);
				break;
		}
		return TRUE;
	}
	else {
		if(_selectedIcon) {
			InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
		}
		_selectedIcon = NULL;
		return FALSE;
	}
}

void Group::_deleteIcons()
{
	for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
		if((*it)->icon) {
			DeleteObject((*it)->icon);
		}
		if((*it)->hbmBuffer) {
			DeleteObject((*it)->hbmBuffer);
		}
		if((*it)->hdcBuffer) {
			DeleteObject((*it)->hdcBuffer);
		}
		free((*it));
	}
	_iconList.clear();
	_iconsLoaded = FALSE;
}

void Group::_loadFolder(const char* szFolder)
{
	// Build file list
	WIN32_FIND_DATA fd;
	HANDLE search;
	char filter[MAX_PATH] = "\0";
	BOOL finished = FALSE;

	// Create file filter
	strcpy(filter, szFolder);
	strcat(filter, "*");

	// Start the search
	search = FindFirstFile(filter, &fd);
	if(search == INVALID_HANDLE_VALUE) {
		LSLogPrintf(LOG_DEBUG, "IconDesk", "No files found.");
	}

	// Loop through the files
	while(!finished) {
		// Don't include "." and ".."
		if(strcmp(fd.cFileName, ".") && strcmp(fd.cFileName, "..")) {
			// Create a icon object
			DESKTOPICON *icon = (DESKTOPICON *)malloc(sizeof(DESKTOPICON));
			memset(icon, 0, sizeof(DESKTOPICON));

			// Save file info
			strcpy(icon->filename, szFolder);
			strcat(icon->filename, fd.cFileName);
			strcpy(icon->title, fd.cFileName);
			if(_iconHideExtension) {
				PathRemoveExtension(icon->title);
			}
			icon->attrib = GetFileAttributes(icon->filename);

			// Hide system files and hidden files
			if(_iconShowHidden || ((icon->attrib & FILE_ATTRIBUTE_HIDDEN) == 0 && (icon->attrib & FILE_ATTRIBUTE_SYSTEM) == 0)) {
				// Add it to the list
				_iconList.push_back(icon);
			}
		}

		// Find next file
		if(!FindNextFile(search, &fd)) {
			if(GetLastError() == ERROR_NO_MORE_FILES) {
				finished = TRUE;
			}
			else {
				LSLogPrintf(LOG_ERROR, "IconDesk", "Couldn't find next file.");
			}
		}
	}

	// Close the search handle.
	if(!FindClose(search)) {
		LSLogPrintf(LOG_ERROR, "IconDesk", "Couldn't close search handle.");
	}
}

void Group::_loadIcons()
{
	for(StringListIter slit = _folders.begin(); slit != _folders.end(); slit++) {
		_loadFolder((*slit).c_str());
	}

	_iconList.sort(Group::_compareIcons);
	if(_iconFoldersFirst) {
		_iconList.sort(Group::_compareIconsFolder);
	}

	// Calculate RECT's and load icons
	int i = 0, x = 0, y = 0;
	IconListIter icon;
	for(icon = _iconList.begin(); icon != _iconList.end(); icon++) {
		switch(_iconDirection) {
			case DD_RIGHT:
			case DD_LEFT:
				x = (i % _iconWrapCount);
				y = (i / _iconWrapCount);
				if(_iconWrapDirection == DD_UP) {
					y = -y;
				}
				if(_iconDirection == DD_LEFT) {
					x = -x;
				}
				break;

			case DD_DOWN:
			case DD_UP:
				x = (i / _iconWrapCount);
				y = (i % _iconWrapCount);
				if(_iconWrapDirection == DD_LEFT) {
					x = -x;
				}
				if(_iconDirection == DD_UP) {
					y = -y;
				}
				break;

		}

		if(!_hideText) {
			(*icon)->x = _iconX + ((max(_iconSize, _iconTextWidth) + _iconSpacingX) * x);
			(*icon)->y = _iconY + ((_iconSize + _iconTextHeight + _iconTextSpacing + _iconSpacingY) * y);

			(*icon)->textRect.top = (*icon)->y + _iconSize + _iconTextSpacing;
			(*icon)->textRect.bottom = (*icon)->textRect.top + _iconTextHeight;
			(*icon)->textRect.left = (*icon)->x;// + (_iconSize / 2) - (_iconTextWidth / 2);
			(*icon)->textRect.right = (*icon)->textRect.left + _iconTextWidth;

			SetRect(&(*icon)->iconRect, (*icon)->x + (max(_iconSize, _iconTextWidth) / 2) - (_iconSize / 2), (*icon)->y, (*icon)->x + (max(_iconSize, _iconTextWidth) / 2) + (_iconSize / 2), (*icon)->y + _iconSize);
		}
		else {
			(*icon)->x = _iconX + ((_iconSize + _iconSpacingX) * x);
			(*icon)->y = _iconY + ((_iconSize + _iconSpacingY) * y);

			SetRect(&(*icon)->iconRect, (*icon)->x, (*icon)->y, (*icon)->x + _iconSize, (*icon)->y + _iconSize);
		}

		(*icon)->icon = _getIconFromPathEx((*icon)->filename, _iconSize, _iconSize);
		if(!(*icon)->icon) {
			LSLogPrintf(LOG_ERROR, "IconDesk", "Could not load icon for \"%s\"", (*icon)->filename);
		}

		// Create buffer copy with fx changes entered

		PAINTSTRUCT ps;
		HDC	hdcScreen = BeginPaint(_parentWnd, &ps);

		if(_effectFlags & FX_ANY) {
			if((*icon)->icon) {
				(*icon)->hdcBuffer = CreateCompatibleDC(hdcScreen);
				(*icon)->hbmBuffer = CreateCompatibleBitmap(hdcScreen, _iconSize, _iconSize);
				SelectObject((*icon)->hdcBuffer, (*icon)->hbmBuffer);

				UINT y = 0;
				HDC hdc = CreateCompatibleDC((*icon)->hdcBuffer);
				HBITMAP hbmNew = CreateCompatibleBitmap((*icon)->hdcBuffer, 2 * _iconSize, _iconSize);
				SelectObject(hdc, hbmNew);

				// Temporarily draw up the icon to read it's pixels
				DrawIconEx(hdc, 0, 0, (*icon)->icon, _iconSize, _iconSize, 0, NULL, DI_NORMAL);
				// the mask of the icon
				DrawIconEx(hdc, _iconSize, 0, (*icon)->icon, _iconSize, _iconSize, 0, NULL, DI_MASK);

				while(y < _iconSize) {
					UINT x = 0;
					while(x < _iconSize) {
						// Loop through all transparent pixels...
						while(x < _iconSize) {
							// If the mask-pixel is white, then break
							if(!GetPixel(hdc, x + _iconSize, y)) {
								COLORREF cl = GetPixel(hdc, x, y);
								BYTE r = GetRValue(cl);
								BYTE g = GetGValue(cl);
								BYTE b = GetBValue(cl);

								// Saturation effect
								if(_effectFlags & FX_SATURATION) {
									BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

									r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
									g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
									b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
								}
								// icon hue effect
								if(_effectFlags & FX_HUE) {
									// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
									// test normal
									r = (BYTE)((r * (255 - _hueIntensity) + GetRValue(_hueColor) * _hueIntensity + 255) >> 8);
									g = (BYTE)((g * (255 - _hueIntensity) + GetGValue(_hueColor) * _hueIntensity + 255) >> 8);
									b = (BYTE)((b * (255 - _hueIntensity) + GetBValue(_hueColor) * _hueIntensity + 255) >> 8);
								}
								SetPixel((*icon)->hdcBuffer, x, y, RGB(r, g, b));
							}
							// for transparent blt?
							else {
								SetPixel((*icon)->hdcBuffer, x, y, RGB(255, 0, 255));
							}
							x++;
						}
					}
					y++;
				}
				DeleteObject(hbmNew);
				DeleteDC(hdc);
			}
		}
		DeleteObject(hdcScreen);
		EndPaint(_parentWnd, &ps);

		i++;
	}

	_iconsLoaded = TRUE;
}

bool Group::_compareIconsFolder(DESKTOPICON* i1, DESKTOPICON* i2)
{
	return ((i1->attrib & FILE_ATTRIBUTE_DIRECTORY) > (i2->attrib & FILE_ATTRIBUTE_DIRECTORY));
};

bool Group::_compareIcons(DESKTOPICON* i1, DESKTOPICON* i2)
{
	return (stricmp(i1->title, i2->title) < 0);
};

void Group::paintIcons(HDC hDC)
{
	if(_visible) {
		if(_iconsLoaded == FALSE) {
			_loadIcons();
		}

		COLORREF oldColor = SetTextColor(hDC, _iconTextColor);
		HFONT oldFont = (HFONT)SelectObject(hDC, _iconTextFont);
		int oldMode = SetBkMode(hDC, TRANSPARENT);

		for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
			if((*it)->icon) {
				if(_effectFlags & FX_ANY) {
					// Paint from the buffer
					TransparentBltLS(hDC, (*it)->x + (max(_iconSize, _iconTextWidth) / 2) - (_iconSize / 2), (*it)->y, _iconSize, _iconSize, (*it)->hdcBuffer, 0, 0, RGB(255, 0, 255));
				}
				else {
					DrawIconEx(hDC, (*it)->x + (max(_iconSize, _iconTextWidth) / 2) - (_iconSize / 2), (*it)->y, (*it)->icon, _iconSize, _iconSize, 0, NULL, DI_NORMAL);
				}
			}
			if(!_hideText) {
				if(_iconTextBackground) {
					FillRect(hDC, &(*it)->textRect, _iconTextBackground);
				}
				if(_selectedIcon == (*it)) {
					DrawFocusRect(hDC, &(*it)->textRect);
				}
				if(_textShadowDepth) {
					RECT r;
					CopyRect(&r, &(*it)->textRect);
					OffsetRect(&r, _textShadowDepth, _textShadowDepth);
					SetTextColor(hDC, _textShadowColor);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, DT_WORDBREAK | DT_CENTER | DT_NOPREFIX /*| DT_NOCLIP*/ | DT_END_ELLIPSIS);
					SetTextColor(hDC, _iconTextColor);
				}
				DrawText(hDC, (*it)->title, strlen((*it)->title), &(*it)->textRect, DT_WORDBREAK | DT_CENTER | DT_NOPREFIX /*| DT_NOCLIP*/ | DT_END_ELLIPSIS);
			}
		}

		SetTextColor(hDC, oldColor);
		SelectObject(hDC, oldFont);
		SetBkMode(hDC, oldMode);
	}
}

DESKTOPICON *Group::_getIconAtPt(int x, int y)
{
	POINT mouse = {x, y};
	for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
		if(PtInRect(&(*it)->textRect, mouse) || PtInRect(&(*it)->iconRect, mouse)) {
			return (*it);
		}
	}
	return NULL;
}

int Group::_parseIconLocation(LPSTR szIconFile)
{
	LPSTR pszComma = strrchr(szIconFile, ',');

	if (!pszComma)
		return -1;

	*pszComma = 0;
	return atoi(pszComma + 1);
}

HICON Group::_getIconFromFile(LPCSTR pszPath, int nIndex, int nSize)
{
	HICON hIcon = NULL;
	HMODULE hLib = LoadLibrary(pszPath);
	if(hLib) {
		hIcon = (HICON)LoadImage(hLib, MAKEINTRESOURCE(nIndex + 1), IMAGE_ICON, nSize, nSize, LR_DEFAULTCOLOR);
		FreeLibrary(hLib);
	}
	return hIcon;
}

HICON Group::_getIconFromPathEx(LPCSTR pszPath, int cxDesired, int cyDesired)
{
	LPCSTR pszExt;

    pszExt = strrchr(pszPath, '.');
        
    if(pszExt && !stricmp(pszExt, ".ico")) {
        return (HICON)LoadImage(NULL, pszPath, IMAGE_ICON, cxDesired, cyDesired, LR_LOADFROMFILE);
    }
    else {
		// Try to extract the icon directly from .exe files
		if(pszExt && !stricmp(pszExt, ".exe")) {
			HICON hIcon = _getIconFromFile(pszPath, 1, cxDesired);
			if(hIcon) {
				return hIcon;
			}
		}

		// Load normal
        SHFILEINFO shfi;
		SHGetFileInfo(pszPath, 0, &shfi, sizeof(SHFILEINFO), SHGFI_ICONLOCATION | SHGFI_ICON | (cxDesired <= 24 ? SHGFI_SMALLICON : 0));

		if(strlen(shfi.szDisplayName)) {
			HICON hIcon = _getIconFromFile(shfi.szDisplayName, shfi.iIcon, cxDesired);
			if(hIcon) {
				DestroyIcon(shfi.hIcon);
				return hIcon;
			}
		}

		return shfi.hIcon;
    }
}

void Group::_getString(const char* key, char* value, const char* defaultValue, int maxLen)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	GetRCString(buffer, value, defaultValue, maxLen);
}

BOOL Group::_getBool(const char* key, BOOL defaultValue)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	return GetRCBoolDef(buffer, defaultValue);
}

int Group::_getInt(const char* key, int defaultValue)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	return GetRCInt(buffer, defaultValue);
}

COLORREF Group::_getColor(const char* key, int defaultValue)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	return GetRCColor(buffer, defaultValue);
}
