/*
This is a part of the IconDesk source code.

Copyright (C) 2003 Erik Christiansson, aka Sci
www.alphafish.com
erik@alphafish.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __GROUP_H
#define __GROUP_H

#include <windows.h>
#include <windowsx.h>
#include <shlwapi.h>
#include <list>
#include <string>
#include "contextmenu.h"
#include "..\lsapi\lsapi.h"
#include "..\lsapi\lswinbase.h"

using namespace std;

struct DESKTOPICON
{
	int x;
	int y;
	RECT iconRect;
	RECT textRect;
	HICON icon;
	char title[MAX_PATH];
	char filename[MAX_PATH];
	DWORD attrib;

	// Icon effects
	HBITMAP hbmBuffer;
	HDC hdcBuffer;
};

enum DESKTOPDIRECTION
{
	DD_RIGHT,
	DD_DOWN,
	DD_LEFT,
	DD_UP
};

typedef list<DESKTOPICON*> IconList;
typedef IconList::iterator IconListIter;

typedef list<string> StringList;
typedef StringList::iterator StringListIter;

typedef list<HANDLE> HandleList;
typedef HandleList::iterator HandleListIter;

// Defined effects
#define FX_HUE			0x1
#define FX_SATURATION	0x2
#define FX_ANY			(FX_HUE | FX_SATURATION)

class Group
{
public:
	Group(const char *name, HWND parentWnd);
	~Group();

	void show();
	void hide();
	void toggle();

	void paintIcons(HDC hDC);
	void update();
	BOOL mouseEvent(UINT uMsg, WPARAM wParam, LPARAM lParam);
	
	//Icons* next;
	char* name;
	//HANDLE iconChangeHandle;
	HandleList changeHandles;

private:
	void _loadIcons();
	void _loadFolder(const char* szFolder);
	void _deleteIcons();

	DESKTOPICON *_getIconAtPt(int x, int y);
	DESKTOPICON *_selectedIcon;

	HICON _getIconFromFile(LPCSTR pszPath, int nIndex, int nSize);
	HICON _getIconFromPathEx(LPCSTR pszPath, int cxDesired, int cyDesired);
	int _parseIconLocation(LPSTR szIconFile);
	void _mouseCommand(DESKTOPICON* clicked, const char* command);

	void _getString(const char* key, char* value, const char* defaultValue, int maxLen);
	BOOL _getBool(const char* key, BOOL defaultValue);
	int _getInt(const char* key, int defaultValue);
	COLORREF _getColor(const char* key, int defaultValue);

	static bool _compareIcons(DESKTOPICON* i1, DESKTOPICON* i2);
	static bool _compareIconsFolder(DESKTOPICON* i1, DESKTOPICON* i2);

	HWND _parentWnd;
	BOOL _visible;

	//DESKTOPICON *_iconList;
	IconList _iconList;
	BOOL _iconsLoaded;

	int _iconX;
	int _iconY;
	int _iconSize;
	int _iconSpacingX;
	int _iconSpacingY;

	StringList _folders;
	//char _iconFolder[MAX_PATH];

	char _mouseLUp[MAX_PATH];
	char _mouseMUp[MAX_PATH];
	char _mouseRUp[MAX_PATH];
	char _mouseLDown[MAX_PATH];
	char _mouseMDown[MAX_PATH];
	char _mouseRDown[MAX_PATH];
	char _mouseLDbl[MAX_PATH];
	char _mouseMDbl[MAX_PATH];
	char _mouseRDbl[MAX_PATH];

	BOOL _iconFoldersFirst;
	BOOL _iconHideExtension;
	BOOL _iconShowHidden;
	BOOL _hideText;

	DESKTOPDIRECTION _iconDirection;
	DESKTOPDIRECTION _iconWrapDirection;
	int _iconWrapCount;

	int _iconTextSpacing;
	int _iconTextHeight;
	int _iconTextWidth;
	HFONT _iconTextFont;
	COLORREF _iconTextColor;
	HBRUSH _iconTextBackground;

	COLORREF _textShadowColor;
	int _textShadowDepth;

	// Icon effects
	COLORREF _hueColor;
	UCHAR _hueIntensity;
	UCHAR _saturation;
	UCHAR _effectFlags;

};

#endif//__GROUP_H
