/*
This is a part of the IconDesk source code.

Copyright (C) 2003 Erik Christiansson, aka Sci
www.alphafish.com
erik@alphafish.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "main.h"

int initModuleEx(HWND hParentWnd, HINSTANCE hInstance, LPCSTR szPath)
{
	desktopWnd = FindWindow("DesktopBackgroundClass", NULL);
	if(!desktopWnd) {
		LSLog(LOG_ERROR, "IconDesk", "Could not find desktop window");
	}

	oldWndProc = (WNDPROC)GetWindowLong(desktopWnd, GWL_WNDPROC);
	SetWindowLong(desktopWnd, GWL_WNDPROC, (LONG)wndProc);
	oldClassStyle = GetClassLong(desktopWnd, GCL_STYLE);
	SetClassLong(desktopWnd, GCL_STYLE, GetClassLong(desktopWnd, GCL_STYLE) | CS_DBLCLKS);

	AddBangCommand("!IconDeskShow", bangShow);
	AddBangCommand("!IconDeskHide", bangHide);
	AddBangCommand("!IconDeskToggle", bangToggle);
	AddBangCommand("!IconDeskUpdate", bangUpdate);

	timerID = GetTickCount();
	SetTimer(desktopWnd, timerID, GetRCInt("IconDeskUpdateInterval", 1000), NULL);

	loadGroups();

	InvalidateRect(desktopWnd, NULL, TRUE);

	return 0;
}

void quitModule(HINSTANCE hInstance)
{
	SetWindowLong(desktopWnd, GWL_WNDPROC, (LONG)oldWndProc);
	SetClassLong(desktopWnd, GCL_STYLE, oldClassStyle);

	RemoveBangCommand("!IconDeskShow");
	RemoveBangCommand("!IconDeskHide");
	RemoveBangCommand("!IconDeskToggle");
	RemoveBangCommand("!IconDeskUpdate");

	KillTimer(desktopWnd, timerID);

	deleteGroups();

	InvalidateRect(desktopWnd, NULL, TRUE);
}

LRESULT wndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg) {
		case WM_LBUTTONDOWN:
		case WM_RBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONUP:
		case WM_MBUTTONUP:
		case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDBLCLK:
		case WM_MBUTTONDBLCLK:
			{
				BOOL hit = FALSE;
				for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
					if((*it)->mouseEvent(uMsg, wParam, lParam)) {
						hit = TRUE;
					}
				}
				if(hit) {
					return 0;
				}
			}
			break;

		case WM_PAINT:
			{
				PAINTSTRUCT	ps;
				HDC			hDC;

				if(!wParam) {
					hDC = BeginPaint(hWnd, &ps);
				}
				else {
					hDC = (HDC)wParam;
				}
				
				for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
					(*it)->paintIcons(hDC);
				}

				if(!wParam) {
					EndPaint(hWnd, &ps);
				}
			}

		case WM_TIMER:
			if(wParam == timerID) {
				for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
					for(list<HANDLE>::iterator chit = (*it)->changeHandles.begin(); chit != (*it)->changeHandles.end(); chit++) {
						if(WaitForSingleObject(*chit, 0) == WAIT_OBJECT_0) {
							(*it)->update();
							InvalidateRect(desktopWnd, NULL, TRUE);
							FindNextChangeNotification(*chit);
						}
					}
				}
				return 0;
			}
			break;
	}
	return CallWindowProc(oldWndProc, hWnd, uMsg, wParam, lParam);
}

void bangShow(HWND caller, LPCSTR args)
{
	Group* group = getGroup(args);
	if(group) {
		group->show();
	}
}

void bangHide(HWND caller, LPCSTR args)
{
	Group* group = getGroup(args);
	if(group) {
		group->hide();
	}
}

void bangToggle(HWND caller, LPCSTR args)
{
	Group* group = getGroup(args);
	if(group) {
		group->toggle();
	}
}

void bangUpdate(HWND caller, LPCSTR args)
{
	Group* group = getGroup(args);
	if(group) {
		group->update();
	}
}

void loadGroups()
{
	char buffer[1024] = "";
	char* group;

	GetRCString("IconDeskGroups", buffer, "", 1024);
	if(strcmp(buffer, "")) {
		group = strtok(buffer, " ,\t");
		while(group) {
			addGroup(group);
			group = strtok(NULL, " ,\t");
		}
	}
	else {
		addGroup("Default");
	}
}

void addGroup(const char* name)
{
	Group* group = new Group(name, desktopWnd);
	groups.push_back(group);
}

void deleteGroups()
{
	for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
		delete *it;
	}
	groups.clear();
}

Group* getGroup(const char* name)
{
	for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
		if(!strcmpi((*it)->name, name)) {
			return *it;
		}
	}
	return NULL;
}
